import { IStateDB, PageConfig, } from '@jupyterlab/coreutils';
import { IMainMenu, } from '@jupyterlab/mainmenu';
import { IDocumentManager, } from '@jupyterlab/docmanager';
import { Menu, } from '@phosphor/widgets';
import { Signal, } from '@phosphor/signaling';
var PluginIDs;
(function (PluginIDs) {
    PluginIDs.recents = 'jupyterlab-recents';
})(PluginIDs || (PluginIDs = {}));
var StateIDs;
(function (StateIDs) {
    StateIDs.recents = `${PluginIDs}:recents`;
})(StateIDs || (StateIDs = {}));
var CommandIDs;
(function (CommandIDs) {
    CommandIDs.openRecent = `${PluginIDs.recents}:open-recent`;
    CommandIDs.clearRecents = `${PluginIDs.recents}:clear-recents`;
})(CommandIDs || (CommandIDs = {}));
var utils;
(function (utils) {
    function mergePaths(root, path) {
        if (root.endsWith('/')) {
            root = root.slice(0, -1);
        }
        if (path.endsWith('/')) {
            path = path.slice(1);
        }
        return `${root}/${path}`;
    }
    utils.mergePaths = mergePaths;
})(utils || (utils = {}));
class RecentsManager {
    constructor(commands, stateDB, contents) {
        this.recentsChanged = new Signal(this);
        // Whether there are local changes sent to be recorded without verification
        this.awaitingSaveCompletion = false;
        this.serverRoot = PageConfig.getOption('serverRoot');
        this.stateDB = stateDB;
        this.contentsManager = contents;
        // This menu will appear in the File menu
        this.recentsMenu = new Menu({ commands });
        this.recentsMenu.title.label = 'Recents';
        // Listen for updates to _recents
        this.recentsChanged.connect((_) => {
            this.syncRecentsMenu();
        });
    }
    get recents() {
        const recents = this._recents || [];
        return recents.filter(r => r.root === this.serverRoot);
    }
    set recents(recents) {
        // Keep track of any recents pertaining to other roots
        const otherRecents = this._recents.filter(r => r.root !== this.serverRoot);
        const allRecents = recents.filter(r => r.root === this.serverRoot).concat(otherRecents);
        this._recents = allRecents;
        this.saveRecents();
        this.recentsChanged.emit(this.recents);
    }
    async init() {
        await this.loadRecents();
        return this.validateRecents();
    }
    addRecent(path, contentType) {
        const recent = {
            root: this.serverRoot,
            path,
            contentType,
        };
        const recents = this.recents;
        const directories = recents.filter(r => r.contentType === 'directory');
        const files = recents.filter(r => r.contentType !== 'directory');
        const destination = contentType === 'directory' ? directories : files;
        // Check if it's already present; if so remove it
        const existingIndex = destination.findIndex(r => r.path === path);
        if (existingIndex >= 0) {
            destination.splice(existingIndex, 1);
        }
        // Add to the front of the list
        destination.unshift(recent);
        // Keep up to 10 of each type of recent path
        if (destination.length > 10) {
            destination.pop();
        }
        this.recents = directories.concat(files);
    }
    removeRecents(paths) {
        const recents = this.recents;
        this.recents = recents.filter(r => paths.indexOf(r.path) === -1);
    }
    clearRecents() {
        this.recents = [];
    }
    async validateRecents() {
        clearTimeout(this.validator);
        // Unless triggered directly, recents will be validated every 12 seconds
        this.validator = setTimeout(this.validateRecents.bind(this), 12 * 1000);
        const recents = this.recents;
        const invalidPathsOrNulls = await Promise.all(recents.map(async (r) => {
            try {
                await this.contentsManager.get(r.path, { content: false });
                return null;
            }
            catch (e) {
                if (e.response.status === 404) {
                    return r.path;
                }
            }
        }));
        const invalidPaths = invalidPathsOrNulls.filter(x => x !== null);
        if (invalidPaths.length > 0) {
            this.removeRecents(invalidPaths);
        }
    }
    syncRecentsMenu() {
        this.recentsMenu.clearItems();
        const recents = this.recents;
        const files = recents.filter(r => r.contentType !== 'directory');
        const directories = recents.filter(r => r.contentType === 'directory');
        [directories, files].forEach(rs => {
            if (rs.length > 0) {
                rs.forEach(recent => {
                    this.recentsMenu.addItem({
                        command: CommandIDs.openRecent,
                        args: { recent },
                    });
                });
                this.recentsMenu.addItem({ type: 'separator' });
            }
        });
        this.recentsMenu.addItem({
            command: CommandIDs.clearRecents,
        });
    }
    async loadRecents() {
        const recents = await this.stateDB.fetch(StateIDs.recents);
        this._recents = recents || [];
        this.recentsChanged.emit(this.recents);
    }
    saveRecents() {
        clearTimeout(this.saveRoutine);
        // Save _recents 500 ms after the last time saveRecents has been called
        this.saveRoutine = setTimeout(async () => {
            // If there's a previous request pending, wait 500 ms and try again
            if (this.awaitingSaveCompletion) {
                this.saveRecents();
            }
            else {
                this.awaitingSaveCompletion = true;
                try {
                    await this.stateDB.save(StateIDs.recents, this._recents);
                    this.awaitingSaveCompletion = false;
                }
                catch (e) {
                    this.awaitingSaveCompletion = false;
                    console.log('Saving recents failed');
                    // Try again
                    this.saveRecents();
                }
            }
        }, 500);
    }
}
const extension = {
    id: PluginIDs.recents,
    autoStart: true,
    requires: [IStateDB, IMainMenu, IDocumentManager],
    activate: (app, stateDB, mainMenu, docManager) => {
        console.log('JupyterLab extension jupyterlab-recents is activated!');
        const { commands, serviceManager } = app;
        ;
        const recentsManager = new RecentsManager(commands, stateDB, serviceManager.contents);
        docManager.activateRequested.connect(async (_, path) => {
            const item = await docManager.services.contents.get(path, {
                content: false
            });
            const fileType = app.docRegistry.getFileTypeForModel(item);
            const contentType = fileType.contentType;
            recentsManager.addRecent(path, contentType);
            // Add the containing directory, too
            if (contentType !== 'directory') {
                const parent = path.lastIndexOf('/') > 0 ? path.slice(0, path.lastIndexOf('/')) : '';
                recentsManager.addRecent(parent, 'directory');
            }
        });
        // Commands
        commands.addCommand(CommandIDs.openRecent, {
            execute: async (args) => {
                const recent = args.recent;
                const path = recent.path === '' ? '/' : recent.path;
                await commands.execute('filebrowser:open-path', { path });
                // If path not found, validating will remove it after an error message
                return recentsManager.validateRecents();
            },
            label: args => {
                const recent = args.recent;
                return utils.mergePaths(recent.root, recent.path);
            },
        });
        commands.addCommand(CommandIDs.clearRecents, {
            execute: () => recentsManager.clearRecents(),
            label: () => 'Clear Recents',
        });
        // Main menu
        mainMenu.fileMenu.addGroup([{
                type: 'submenu',
                submenu: recentsManager.recentsMenu,
            }], 1);
        // Try to merge with existing Group 1
        try {
            const groups = mainMenu.fileMenu._groups;
            let numRankOneGroups = 0;
            let openGroupIndex = -1;
            for (let i = 0; i < groups.length; i++) {
                const group = groups[i];
                if (group.rank === 1) {
                    numRankOneGroups += 1;
                    if (openGroupIndex < 0) {
                        openGroupIndex = i;
                    }
                }
            }
            if (numRankOneGroups === 2) {
                const openGroup = groups[openGroupIndex];
                openGroup.size = openGroup.size + 1;
                groups.splice(openGroupIndex + 1, 1);
                const fileMenu = mainMenu.fileMenu.menu;
                const fileMenuItems = fileMenu._items;
                let removeSeparators = false;
                for (let i = fileMenuItems.length - 1; i > 0; i--) {
                    const fileMenuItem = fileMenuItems[i];
                    if (fileMenuItem.command === 'filebrowser:open-path') {
                        break;
                    }
                    if (removeSeparators && fileMenuItem.type === 'separator') {
                        fileMenu.removeItemAt(i);
                    }
                    else if (fileMenuItem.type === 'submenu') {
                        const label = fileMenuItem.submenu.title.label;
                        if (label === 'Recents') {
                            removeSeparators = true;
                        }
                    }
                }
            }
        }
        catch (e) { }
        recentsManager.init();
    }
};
export default extension;
